# encoding: utf-8

'''☎️ EDRN LDAP Sync: Resource Description Format classes and functions.'''


import rdflib, logging, enum
from dataclasses import dataclass

_logger = logging.getLogger(__name__)


# RDF Predicates
# --------------

_account_pred_uri  = rdflib.term.URIRef('http://xmlns.com/foaf/0.1/accountName')
_chair_pred_uri    = rdflib.term.URIRef('http://edrn.nci.nih.gov/xml/rdf/edrn.rdf#chair')
_cochair_pred_uri  = rdflib.term.URIRef('http://edrn.nci.nih.gov/xml/rdf/edrn.rdf#coChair')
_com_type_pred_uri = rdflib.term.URIRef('http://edrn.nci.nih.gov/xml/rdf/edrn.rdf#committeeType')
_email_pred_uri    = rdflib.term.URIRef('http://xmlns.com/foaf/0.1/mbox')
_gn_pred_uri       = rdflib.term.URIRef('http://xmlns.com/foaf/0.1/givenname')
_member_pred_uri   = rdflib.term.URIRef('http://edrn.nci.nih.gov/xml/rdf/edrn.rdf#member')
_phone_pred_uri    = rdflib.term.URIRef('http://xmlns.com/foaf/0.1/phone')
_pi_pred_uri       = rdflib.term.URIRef('http://edrn.nci.nih.gov/rdf/schema.rdf#pi')
_sn_pred_uri       = rdflib.term.URIRef('http://xmlns.com/foaf/0.1/surname')
_staff_pred_uri    = rdflib.term.URIRef('http://edrn.nci.nih.gov/rdf/schema.rdf#staff')
_status_pred_uri   = rdflib.term.URIRef('http://edrn.nci.nih.gov/rdf/schema.rdf#employmentActive')


# RDF object types
# ----------------

_site_type = rdflib.term.URIRef('http://edrn.nci.nih.gov/rdf/types.rdf#Site')
_committee_type = rdflib.term.URIRef('http://edrn.nci.nih.gov/rdf/types.rdf#Committee')


# Classes & Enums
# ---------------

class Status(enum.Enum):
    '''The status of an EDRN person.'''
    UNKNOWN = -1
    INACTIVE = 0
    ACTIVE = 1


@dataclass(order=True, frozen=True)
class Person(object):
    '''And EDRN person.'''
    uid: str
    sn: str
    gn: str
    email: str
    phone: str
    status: Status
    subject_uri: str

    @property
    def cn(self) -> str:
        return f'{self.gn} {self.sn}'


@dataclass(order=True, frozen=True)
class Group(object):
    '''A geneeric group (could be a site or a committee) with a common name and members.'''
    cn: str
    members: set


class Site(Group):
    '''An EDRN funded site.'''
    pass


class Committee(Group):
    '''An EDRN committee.'''
    pass


# Functions
# ---------

def parse(url: str):
    '''Parse the given RDF ``url`` and return a dict of statements (subject URIs to predicates), which itself is a dict
    of predicate URIs to sequences of objects.
    '''
    statements = {}
    graph = rdflib.Graph()
    graph.parse(url)
    for s, p, o in graph:
        predicates = statements.get(s, dict())
        objects = predicates.get(p, list())
        objects.append(o)
        predicates[p] = objects
        statements[s] = predicates
    _logger.debug('Read %d statements from %s', len(statements), url)
    return statements


def sv(predicate_uri: rdflib.term.URIRef, predicates: dict) -> object:
    '''From the dict of ``predicates`` return the first value only matching ``predicate_uri`` or ``None``
    if the predicate doesn't occur within the ``predicates``. Here, ``sv`` means "single value".
    '''
    return predicates.get(predicate_uri, [None])[0]


def get_rdf_type(predicates: dict) -> rdflib.term.URIRef:
    '''Get the single RDF type of the object described by the given ``predicates``, or ``None`` if it's
    not available.
    '''
    return predicates.get(rdflib.RDF.type, [None])[0]


def get_rdf_people(url: str) -> dict:
    '''Return all people currently in RDF as a mapping from user ID to ``Person`` objects. Skip anyone who is
    without an account name.
    '''
    statements, people = parse(url), {}
    for subject_uri, predicates in statements.items():
        uid = sv(_account_pred_uri, predicates)
        if not uid: continue
        uid = str(uid).lower()

        status = sv(_status_pred_uri, predicates)
        if not status:
            status = Status.UNKNOWN
        elif str(status) == 'Former employee':
            status = Status.INACTIVE
        else:
            status = Status.ACTIVE

        sn = sv(_sn_pred_uri, predicates)
        sn = str(sn) if sn else 'UNKNOWN'
        gn = sv(_gn_pred_uri, predicates)
        gn = str(gn) if gn else 'UNKNOWN'

        email = sv(_email_pred_uri, predicates)
        if not email:
            email = 'UNKNOWN@UNKNOWN.COM'
        else:
            # Strip the 'mailto:'
            email = str(email[7:])

        p = Person(str(uid), str(sn), str(gn), email, str(sv(_phone_pred_uri, predicates)), status, subject_uri)
        people[uid] = p

    return people


def get_rdf_sites(rdf_people: dict, url: str) -> dict:
    '''Return all sites currently in RDF as a mapping from site name (which is the bizarre convention of
    the principal investigator's last name plus the name of the institution) to ``Site`` objects. We skip
    any site for which we cannot determine a PI. This takes as input the ``rdf_people`` output from
    ``get_rdf_people`` and the ``url`` to the site RDF data source.
    '''
    people = {i.subject_uri: i for i in rdf_people.values()}
    statements, sites = parse(url), {}
    for predicates in statements.values():
        if get_rdf_type(predicates) != _site_type: continue     # Not a Site? Skip it
        title = sv(rdflib.DCTERMS.title, predicates)            # Grab thte name of the site
        title = str(title).strip().replace(',', '')             # DMCC data is always weirdly padded; see [1]
        if not title: continue                                  # Unnamed? Skipt it
        pi = people.get(sv(_pi_pred_uri, predicates))           # Find the PI
        if not pi: continue                                     # No principal investigator? Skip it
        site_name = f'{pi.sn} {title}'                          # Figure out weird site name we use in EDRN
        members = set([pi])                                     # Start with the pi
        for person_uri in predicates.get(_staff_pred_uri, []):  # For each member's RDF subject URI
            member = people.get(person_uri)                     # Find the member
            if member: members.add(member)                      # If we got one, note it
        site = Site(site_name, members)
        sites[site_name] = site
    return sites


def get_rdf_collab_groups(rdf_people: dict, url: str) -> dict:
    '''Get the collaborative groups (only) from the committee RDF source at ``url`` and filling each
    group with people from ``rdf_people``, returning a dict of committee CN to ``Committee``.
    '''
    people = {i.subject_uri: i for i in rdf_people.values()}
    statements, groups = parse(url), {}
    for predicates in statements.values():
        if get_rdf_type(predicates) != _committee_type: continue
        if str(sv(_com_type_pred_uri, predicates)).strip() != 'Collaborative Group': continue
        title = sv(rdflib.DCTERMS.title, predicates)
        title = str(title).strip()
        if not title: continue
        title = title[0:title.rindex(' Cancers Research Group')]
        members = set()
        chair = people.get(sv(_chair_pred_uri, predicates))
        if chair: members.add(chair)
        cochair = people.get(sv(_cochair_pred_uri, predicates))
        if cochair: members.add(cochair)
        for person_uri in predicates.get(_member_pred_uri, []):
            member = people.get(person_uri)
            if member: members.add(member)
        group = Committee(title, members)
        groups[title] = group
    return groups


# [1] While the common name for some sites may indeed be "Boutros The University of California, Los Angeles",
# the best practices for LDAP (see https://cutt.ly/aBNFeUE) suggest avoiding any special characaters in
# naming attributes in RDNs. Chris Mattmann established the practice of replacing commas with spaces in
# an earlier version of this software, but this left a group named with an awkward double-space. Here, we
# elect to just drop commas.



# encoding: utf-8

'''☎️ EDRN LDAP Sync: constants.'''

import ldap


# LDAP-related
# ------------

ALL_GROUP_DN         = 'cn=All EDRN,dc=edrn,dc=jpl,dc=nasa,dc=gov'
COMMITTEE_RDF_URL    = 'https://edrn.jpl.nasa.gov/cancerdataexpo/rdf-data/committees/@@rdf'
GROUP_BASE           = 'dc=edrn,dc=jpl,dc=nasa,dc=gov'
GROUP_OBJECT_CLASSES = ['groupOfUniqueNames', 'top']
LDAP_URL             = 'ldaps://edrn-ds.jpl.nasa.gov'
MANAGER_DN           = 'uid=admin,ou=system'
SITE_RDF_URL         = 'https://edrn.jpl.nasa.gov/cancerdataexpo/rdf-data/sites/@@rdf'
USER_BASE            = 'dc=edrn,dc=jpl,dc=nasa,dc=gov'
USER_CLASS           = 'edrnPerson'
USER_OBJECT_CLASSES  = [USER_CLASS, 'inetOrgPerson', 'organizationalPerson', 'person', 'top']
USER_RDF_URL         = 'https://edrn.jpl.nasa.gov/cancerdataexpo/rdf-data/registered-person/@@rdf'


# Map from command-line to ldap constants
# ---------------------------------------

LDAP_SCOPES = {
    'base': ldap.SCOPE_BASE,
    'one': ldap.SCOPE_ONELEVEL,
    'sub': ldap.SCOPE_SUBTREE
}



# encoding: utf-8

'''☎️ EDRN LDAP Sync: group sync with data from the DMCC.'''

from . import VERSION
from .constants import USER_RDF_URL, SITE_RDF_URL, COMMITTEE_RDF_URL, GROUP_BASE, GROUP_OBJECT_CLASSES, USER_BASE
from .rdf import get_rdf_people, get_rdf_sites, get_rdf_collab_groups, Group
from .utils import add_logging_arguments, add_ldap_arguments, get_manager_password
import sys, ldap, ldap.modlist, logging, argparse

__version__ = VERSION
_logger = logging.getLogger(__name__)
_prog_desc = 'Synchronizes EDRN groups from the DMCC SOAP (RDF) feed into the EDRN LDAP'


def _add_to_ldap(connection: ldap.ldapobject.LDAPObject, group: Group):
    '''Add to the LDAP directory at the given ``connection`` the ``group``.'''
    dn = f'cn={group.cn},{GROUP_BASE}'
    name = group.cn.encode('utf-8')
    members = [f'uid={i.uid},{USER_BASE}'.encode('utf-8') for i in group.members]
    attrs = {
        'objectClass': [i.encode('utf-8') for i in GROUP_OBJECT_CLASSES],
        'cn': name,
        'description': name,
        'uniqueMember': members
    }
    modlist = ldap.modlist.addModlist(attrs)
    try:
        _logger.debug('Trying to add group %s', dn)
        connection.add_s(dn, modlist)
    except ldap.ALREADY_EXISTS:
        _logger.debug('%s already exists, so modifying its members', dn)
        mods = [(ldap.MOD_REPLACE, 'uniqueMember', members)]
        connection.modify_s(dn, mods)


def main():
    '''Here we go. Daddy.'''
    parser = argparse.ArgumentParser(description=_prog_desc)
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')
    add_logging_arguments(parser)
    add_ldap_arguments(parser)
    parser.add_argument('--user-rdf-url', default=USER_RDF_URL, help='URL to RDF info about people [%(default)s]')
    parser.add_argument('--site-rdf-url', default=SITE_RDF_URL, help='URL to RDF info about sites [%(default)s]')
    parser.add_argument('--committee-rdf-url', default=COMMITTEE_RDF_URL, help='URL to RDF info about committees [%(default)s]')

    args = parser.parse_args()
    logging.basicConfig(level=args.loglevel, format='%(levelname)s %(message)s')

    rdf_people = get_rdf_people(args.user_rdf_url)
    _logger.info('Users found in DMCC RDF: %d', len(rdf_people))
    rdf_sites = get_rdf_sites(rdf_people, args.site_rdf_url)
    _logger.info('Sites found in DMCC RDF: %d', len(rdf_sites))
    rdf_collab_groups = get_rdf_collab_groups(rdf_people, args.committee_rdf_url)
    _logger.info('Collab Groups found in DMCC RDF: %d', len(rdf_collab_groups))

    connection = ldap.initialize(args.url)
    password = get_manager_password(args)
    connection.simple_bind_s(args.manager_dn, password)

    for group in rdf_sites.values():
        _add_to_ldap(connection, group)
    for group in rdf_collab_groups.values():
        _add_to_ldap(connection, group)

    sys.exit(0)



# encoding: utf-8

'''☎️ EDRN LDAP Sync: user sync with data from the DMCC.'''

from . import VERSION
from .constants import USER_BASE, USER_RDF_URL, USER_OBJECT_CLASSES
from .rdf import get_rdf_people, Person, Status
from .utils import add_logging_arguments, add_ldap_arguments, get_manager_password
import sys, ldap, ldap.modlist, logging, datetime, random, string, argparse

_logger = logging.getLogger(__name__)
_prog_desc = 'Synchronizes EDRN users from the DMCC SOAP (RDF) feed into the EDRN LDAP'
_password_corpus = string.ascii_letters + string.digits + string.punctuation
__version__ = VERSION


def _timestamp() -> str:
    '''Make a UTC timestamp as an ISO 8601 string.'''
    return datetime.datetime.now(datetime.timezone.utc).isoformat()


def _generate_description() -> str:
    '''Make a suitable description of a person, which is just a timestamp for our purposes.'''
    return 'imported via EDRN dmccsync at ' + _timestamp()


def _generate_password() -> str:
    '''Make a random password.'''
    return '{CRYPT}' + ''.join(random.sample(_password_corpus, 16))


def _add(person: Person, connection: ldap.ldapobject.LDAPObject):
    '''Add the ``person`` to the LDAP server at ``connection``.'''
    dn = f'uid={person.uid},{USER_BASE}'
    attrs = {
        'cn': person.cn.encode('utf-8'),
        'description': _generate_description().encode('utf-8'),
        'mail': person.email.encode('utf-8'),
        'objectClass': [i.encode('utf-8') for i in USER_OBJECT_CLASSES],
        'sn': person.sn.encode('utf-8'),
        'telephoneNumber': person.phone.encode('utf-8'),
        'uid': person.uid.encode('utf-8'),
        'userPassword': _generate_password().encode('utf-8'),
    }
    modlist = ldap.modlist.addModlist(attrs)
    try:
        _logger.debug('Add new user %s', person.uid)
        connection.add_s(dn, modlist)
    except ldap.ALREADY_EXISTS:
        pass


def main():
    '''Let's do this.'''
    parser = argparse.ArgumentParser(description=_prog_desc)
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')
    add_logging_arguments(parser)
    add_ldap_arguments(parser)
    parser.add_argument('--userbase', default=USER_BASE, help='Base DN where users are found [%(default)s]')
    parser.add_argument('--user-rdf-url', default=USER_RDF_URL, help='URL to RDF information about people')

    args = parser.parse_args()
    logging.basicConfig(level=args.loglevel, format='%(levelname)s %(message)s')

    rdf_people = get_rdf_people(args.user_rdf_url)
    _logger.info('Users found in DMCC RDF: %d', len(rdf_people))
    active_uids, inactive_uids = set(), set()
    for person in rdf_people.values():
        if person.status == Status.ACTIVE:
            active_uids.add(person.uid)
        else:
            inactive_uids.add(person.uid)
    _logger.info('Of those, the actives number %d and the inactives %d', len(active_uids), len(inactive_uids))

    connection = ldap.initialize(args.url)
    password = get_manager_password(args)
    connection.simple_bind_s(args.manager_dn, password)
    ldap_people = connection.search_s(args.userbase, ldap.SCOPE_ONELEVEL, '(objectClass=edrnPerson)', ['uid'])
    ldap_uids = set([i[1]['uid'][0].decode('utf-8') for i in ldap_people])
    _logger.info('Number of edrnPersons in LDAP: %d', len(ldap_uids))

    # Delete all inactives
    dead_uids = inactive_uids & ldap_uids
    for dead_uid in dead_uids:
        try:
            _logger.debug('Deleting user %s', dead_uid)
            connection.delete_s(f'uid={dead_uid},{USER_BASE}')
        except ldap.NO_SUCH_OBJECT:
            pass
    _logger.info('Count of inactive users deleted: %d', len(dead_uids))

    # Add in whoever's new to LDAP
    new_uids = active_uids - ldap_uids
    for new_uid in new_uids:
        person = rdf_people[new_uid]
        _add(person, connection)
    _logger.info('Count of new users added: %d', len(new_uids))

    sys.exit(0)


if __name__ == '__main__':
    main()



# encoding: utf-8

'''☎️ EDRN LDAP Sync: utilities.'''

from .constants import LDAP_URL, MANAGER_DN
import logging, argparse, os, getpass


def add_logging_arguments(parser: argparse.ArgumentParser):
    '''Add command-line argument support to ``parser`` for logging controls.'''
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--debug', action='store_const', dest='loglevel', const=logging.DEBUG, default=logging.INFO,
        help='🐞 Log copious and verbose messages suitable for developers'
    )
    group.add_argument(
        '--quiet', action='store_const', dest='loglevel', const=logging.WARNING,
        help="🤫 Don't log informational messages"
    )


def add_ldap_arguments(parser: argparse.ArgumentParser):
    '''Add command-line arguments to support identifying and connecting to an LDAP server.'''
    group = parser.add_argument_group('LDAP Server', description='How to identify and authenticate with LDAP')
    group.add_argument('-H', '--url', default=LDAP_URL, help='URL to the LDAP server [%(default)s]')
    group.add_argument('-D', '--manager-dn', default=MANAGER_DN, help='DN of the manager user [%(default)s]')
    group.add_argument(
        '-w', '--password',
        help='Password for the manager DN; defaults to MANAGER_DN_PASSWORD env var, or will be prompted if unset'
    )


def get_manager_password(options: argparse.Namespace) -> str:
    '''Get the LDAP's manager password from the given ``options``. If it's not there, then try the
    MANAGER_DN_PASSWORD environment variable. And if that's not given, then prompt for it.
    '''
    password = options.password
    if not password:
        password = os.getenv('MANAGER_DN_PASSWORD')
        if not password:
            password = getpass.getpass(f'Password for {options.manager_dn}: ')
    return password



# encoding: utf-8

'''☎️ Keep the ``All EDRN`` group up-to-date.'''

from . import VERSION
from .constants import ALL_GROUP_DN as GROUP_DN
from .constants import LDAP_SCOPES
from .constants import USER_BASE, USER_CLASS
from .utils import add_logging_arguments, add_ldap_arguments, get_manager_password
import sys, ldap, argparse, logging

__version__ = VERSION
_logger = logging.getLogger(__name__)
_prog_desc = 'Synchronizes the `All EDRN` group in EDRN LDAP'
SCOPE = 'one'


def main():
    '''Here we go. Rooster sound.'''
    parser = argparse.ArgumentParser(description=_prog_desc)
    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')
    add_logging_arguments(parser)
    add_ldap_arguments(parser)
    parser.add_argument('--userbase', default=USER_BASE, help='Base DN where users are found [%(default)s]')
    parser.add_argument(
        '-s', '--scope', default=SCOPE, choices=['base', 'one', 'sub'],
        help='Search scope to find users [%(default)s]'
    )
    parser.add_argument('--userclass', default=USER_CLASS, help='Object class expected of users [%(default)s]')
    parser.add_argument('--group', default=GROUP_DN, help='DN of the "all users" group to update [%(default)s]')

    args = parser.parse_args()
    logging.basicConfig(level=args.loglevel, format='%(levelname)s %(message)s')

    userbase, scope, userclass, group = args.userbase, args.scope, args.userclass, args.group
    connection = ldap.initialize(args.url)
    password = get_manager_password(args)
    connection.simple_bind_s(args.manager_dn, password)

    all_users = connection.search_s(userbase, LDAP_SCOPES[scope], '(objectClass={})'.format(userclass), [], attrsonly=1)
    all_users = set([i[0] for i in all_users])

    current_members = connection.search_s(group, ldap.SCOPE_BASE, '(objectClass=*)', ['uniqueMember'])
    try:
        # OpenLDAP returns the key ``uniqueMember`` with capital ``M``
        current_members = set([str(i, 'utf-8') for i in current_members[0][1]['uniqueMember']])
    except KeyError:
        # Apache DS returns the key ``uniquemember`` with lowercase ``m``
        current_members = set([str(i, 'utf-8') for i in current_members[0][1]['uniquemember']])

    to_add = all_users - current_members
    to_remove = current_members - all_users
    if to_add:
        _logger.debug('Adding these people to %s: %r', to_add, group)
        connection.modify_s(group, [(ldap.MOD_ADD, 'uniqueMember', [i.encode('utf-8') for i in to_add])])
    if to_remove:
        _logger.debug('Removing these people from %s: %r', to_remove, group)
        connection.modify_s(group, [(ldap.MOD_DELETE, 'uniqueMember', [i.encode('utf-8') for i in to_remove])])

    sys.exit(0)


if __name__ == '__main__':
    main()
